%{
/*
 * Copyright (c) 2000-2003 QoSient, LLC
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*
 * Copyright (c) 1988, 1989, 1990, 1991, 1992, 1993, 1994
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */



#include <ctype.h>
#include <string.h>
#include <compat.h>

#include <pcap.h>
#include <argus-namedb.h>

#include <gencode.h>
#include <tokdefs.h>
#include <syslog.h>

extern void ArgusLog (int, char *, ...);

#ifndef __GNUC__
#define inline
#endif

int argus_lex(void);
static int stoi(char *);
static inline int xdtoi(int);

#ifdef FLEX_SCANNER
#undef YY_INPUT
#define YY_INPUT(buf, result, max)\
{\
   char *src = in_buffer;\
   int i;\
\
   if (*src == 0)\
      result = YY_NULL;\
   else {\
      for (i = 0; *src && i < max; ++i)\
         buf[i] = *src++;\
      in_buffer += i;\
      result = i;\
   }\
}
#else
#undef getc
#define getc(fp)  (*in_buffer == 0 ? EOF : *in_buffer++)
#endif

#define yylval argus_lval
extern YYSTYPE yylval;

static char *in_buffer = NULL;

%}

N		([0-9]+|(0X|0x)[0-9A-Fa-f]+)
B		([0-9A-Fa-f][0-9A-Fa-f]?)

%a 3000

%%
dst		return DST;
src		return SRC;

link|ether|ppp|slip  return LINK;
fddi		return LINK;
arp		return ARP;
rarp		return RARP;
ip		return IP;
tcp		return TCP;
udp		return UDP;
rtp		return RTP;
rtcp		return RTCP;
esp		return ESP;
man		return MAN;
icmp		return ICMP;
igmp		return IGMP;
mpls		return MPLS;
vlan		return VLAN;
vid		return VLANID;
mid		return MPLSID;
anon		return ANON;

test		return TEST;
host		return HOST;
srcid		return SRCID;
net		return NET;
mask		return MASK;
port		return PORT;
proto		return PROTO;

ttl		return TTL;
tos		return TOS;

gateway		return GATEWAY;

less		return LESS;
greater		return GREATER;
byte		return BYTE;
broadcast	return TK_BROADCAST;
multicast	return TK_MULTICAST;

and|"&&"	return AND;
or|"||"		return OR;
not		return '!';

len|length	return LEN;
inbound		return INBOUND;
outbound	return OUTBOUND;

syn		return SYN;
synack		return SYNACK;
data		return DATA;
fin		return FIN;
finack		return FINACK;

rtr		return RTR;
mbr		return MBR;
lvg		return LVG;

reset		return RESET;
normal		return NORMAL;
wait		return WAIT;
timeout		return TIMEDOUT;
est		return ESTABLISHED;
con		return CONNECTED;
cor		return CORRELATED;
corolated	return CORRELATED;

ecn		return ECN;
drop		return RETRANS;
retrans		return RETRANS;
outoforder	return OUTOFORDER;
ooo		return OUTOFORDER;
multipath	return MULTIPATH;
winshut		return WINSHUT;
frag		return FRAG;
fragonly	return FRAG_ONLY;

echo		return ECHO;
unreach		return UNREACH;
redirect	return REDIRECT;
timexed		return TIMEXED;

[ \n\t]			;
[+\-*/:\[\]!<>()&|=]	return yytext[0];
">="			return GEQ;
"<="			return LEQ;
"!="			return NEQ;
"=="			return '=';
"<<"			return LSH;
">>"			return RSH;
{N}			{ yylval.i = stoi((char *)yytext); return NUM; }
({N}\.{N})|({N}\.{N}\.{N})|({N}\.{N}\.{N}\.{N})	{
			yylval.s = Argussdup((char *)yytext); return HID; }
{B}:{B}:{B}:{B}:{B}:{B} { yylval.e = argus_ether_aton((char *)yytext);
			  return EID; }
{B}:+({B}:+)+		{ ArgusLog(LOG_ERR, "bogus ethernet address %s", yytext); }
[A-Za-z0-9][-_.A-Za-z0-9]*[.A-Za-z0-9] {
			yylval.s = Argussdup((char *)yytext); return ID; }
"\\"[^ !()\n\t]+	{ yylval.s = Argussdup((char *)yytext + 1); return ID; }
[^ \[\]\t\n\-_.A-Za-z0-9!<>()&|=]+i {
			ArgusLog(LOG_ERR, "illegal token: %s\n", yytext); }
.			{ ArgusLog (LOG_ERR, "illegal char '%c'", *yytext); }
%%
void
argus_lex_init(buf)
char *buf;
{
   in_buffer = strdup(buf);
}

/*
 * Also define a yywrap.  Note that if we're using flex, it will
 * define a macro to map this identifier to pcap_wrap.
 */
int
yywrap()
{
	return 1;
}

/* Hex digit to integer. */
static inline int
xdtoi(c)
int c;
{
   if (isdigit(c))
      return c - '0';
   else if (islower(c))
      return c - 'a' + 10;
   else
      return c - 'A' + 10;
}

/*
 * Convert string to integer.  Just like atoi(), but checks for
 * preceding 0x or 0 and uses hex or octal instead of decimal.
 */
static int
stoi(s)
	char *s;
{
	int base = 10;
	int n = 0;

	if (*s == '0') {
		if (s[1] == 'x' || s[1] == 'X') {
			s += 2;
			base = 16;
		}
		else {
			base = 8;
			s += 1;
		}
	}
	while (*s)
		n = n * base + xdtoi(*s++);

	return n;
}

